{
        *****  CPAS Portable Shell External Environment  *******

*************************************************************************
*                                                                       *
*                                                                       *
*  F I T - L S Q  (Library environment perform any least-Squares Fit )  *
*                                                                       *
*                                                                       *
*                Version  1.1-C  - - 30-Nov-2009                        *
*                                                                       *
*                                by                                     *
*                                                                       *
*                  Pierre Wolfers, Institut Neel                        *
*                                                                       *
*          CNRS GRENOBLE,  25 Avenue des Martyrs, B.P. 166              *
*                                                                       *
*                     F 38042 GRENOBLE CEDEX 9                          *
*                                                                       *
*                           F R A N C E                                 *
*                                                                       *
*                                                                       *
*                                                                       *
*************************************************************************

/////////////////////////////////////////////////////////////////////////
//                                                                     //
//                                                                     //
//                  Global Public Licence (GPL)                        //
//                                                                     //
//                                                                     //
// This license described in this file overrides all other licenses    //
// that might be specified in other files for this library.            //
//                                                                     //
// This library is free software; you can redistribute it and/or       //
// modify it under the terms of the GNU Lesser General Public          //
// License as published by the Free Software Foundation; either        //
// version 2.1 of the License, or (at your option) any later version.  //
//                                                                     //
// This library is distributed in the hope that it will be useful,     //
// but WITHOUT ANY WARRANTY; without even the implied warranty of      //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU   //
// Library General Public License for more details.                    //
//                                                                     //
// You should have received a copy of the GNU Lesser General Public    //
// License along with this library (see COPYING.LIB); if not, write to //
// the Free Software Foundation :                                      //
//                      Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

}


  {**********************************************************************************
  *                                                                                 *
  *                        Global  Constants  Declarations                          *
  *                                                                                 *
  **********************************************************************************}


const
  flt_max       =              1.0e+30;                 { Define the largest floatting number constant }


type
  lsq_flt       =               double;                 { For Least-Squares process }


[external]
var

  ndat,                                                 { Number of data }
  np,                                                   { Total number of parameters }
  nsing:                       integer;                 { Number of detected singularities }

  min_diag:                    lsq_flt;                 { Minimum for a pivot to avoid a singularity }

  chi2eps,                                              { Increasing tolerance for chi2 in Marquardt mode }
  chi2,                                                 { Chi_squared = Goodness of fit }
  lchi2,                                                { Previous Chi_squared }
  nchi2:                       lsq_flt;                 { Chi_squared*(n-v) = Goodness of fit numerator }




{ *** Least Squares procedures/functions *** }


{ Initialize routine. Can be used to re-int the Least-Squares Module }
procedure LSQ$INIT; external;

{ Define a parameter to FIT
  prm and sig are the parameter and its related standard error,
  min and max are the minimum and maximum allowed of prm.
  LSQ$NEW_VPARM return the <p_id> integer identifier of the parameter to fit.
}
function  LSQ$NEW_VPARM( var prm, sig: lsq_flt; min, max: lsq_flt := flt_max ): integer; external;

{ Function to organize the previously definy parameters (by LSQ$NEW_VPARM) in a Least-Squares Block
  where dmp and mrq are respectively the damping and Levenberg-Marquardt factors (in the range ]0.0,1.0].
  This function return the integer identifier of the new Least-Squares block.
}
function  LSQ$NEW_BLOCK( dmp, mrq: lsq_flt ): integer; external;

{ Procedure to modify the damping (dmp) and Levenberg-Marquardt (mrq) factors during the fit.
  Can be called only between to cycles.
  id   is the related Least-Squares block as returned by LSQ$NEW_BLOCK.
}
procedure LSQ$SET_DMP_MRQ( id: integer; dmp, mrq: lsq_flt ); external;

{ Procedure to initilize a new Least-Sqaures cycle. }
procedure LSQ$INIT_CYCLE; external;

{ Set an observation contribution during a cycle.
  delta   is the difference observed value - comuted value,
  weight  is the related weight (standard mode is 1/sigma - internaly squared),
  drvtab  is an array off all partial derivate (from each parameter) of the computed value.
          The size og drvtab must be equal to the total number of parameters to fit.
}
procedure LSQ$SET_ELEM( delta, weight: lsq_flt; in_var drvtab: array[size: integer] of lsq_flt );
external;

{ The three following procedures make an alternative to use LSQ$SET_ELEM.
  LSQ$INIT_DERV  init the vector of partial derivate.
  LSQ$SET_DERV   set the partial derivate d<obs>/dParam<p_id> where <p_id> is the
                 identifier number of a parameter to fit.
  LSQ$SET_CONTR  complete the measure contribution.
                 delta and weight are same that with LSQ$SET_ELEM.
                 any missing call of LSQ$SET_DERV set the related derivate to 0.0 .
}
procedure LSQ$INIT_DERV; external;

procedure LSQ$SET_DERV( id: integer; drv: lsq_flt ); external;

procedure LSQ$SET_CONTR( delta, weight: lsq_flt ); external;

{ Function to resume one Least-Squares cycle where :

 bend      when true, computes the correlation matrix. 
 bchg      if true, all values of parameters to fit will be changed by call of the user function CH_PARM.
 bmrq      if true, any cycle with <chi2> > the <previous_cycle_chi2> do not change the parameter values,
 
 CH_PARM   is a user function called by LSQ$PROCESS for each parameter to fit (with respect of bchg and bmrq conditions).
           the related arguments are :
     v_o   The old value of the current parameter,
     v_n   its new proposed value (as computed frm least-squares algorithm,
     v_c   the value change (= v_n - v_o),
     v_s   the new standard error of the current parameter,
     min   minimum value allowed for the current parameter,
     max   minimum value allowed for the current parameter.
 
     This function must be return the new value to set in the current parameter to fit.

  The function LSQ$PROCESS return the value of bchg or false when bmrq was true and the chi2 is increasing. 
}
function  LSQ$PROCESS( bend, bchg, bmrq: boolean;
          function CH_PARM( i: integer; v_o, v_n, v_c, v_s, min, max: lsq_flt ): lsq_flt ): boolean;
external;

{ procedure is used to restore the values of each parameter to fit of the previous cycle.
  With this procedure it is possible re-compute the last cycle with some other Least_Squares block
}
procedure LSQ$RESTORE_PARM; external;

{ Function to get the correlation factor between two parameter to fit.
  ii and jj are the two integer identifiers as return by LSQ$NEW_VPARM.
}
function  LSQ$CORREL( ii, jj: integer ): lsq_flt;
external;

{ *** End of environment file *** }
