{
*************************************************************************
*                                                                       *
*                                                                       *
*                        F I T S H I P                                  *
*                                                                       *
*                                                                       *
*                Version  1.0-A  - - 14-Oct-2014                        *
*                                                                       *
*                                by                                     *
*                                                                       *
*                  Pierre Wolfers, Institut Neel                        *
*                                                                       *
*          CNRS GRENOBLE,  25 Avenue des Martyrs, B.P. 166              *
*                                                                       *
*                     F 38042 GRENOBLE CEDEX 9                          *
*                                                                       *
*                           F R A N C E                                 *
*                                                                       *
*                                                                       *
*                                                                       *
*************************************************************************

/////////////////////////////////////////////////////////////////////////
//                                                                     //
//                                                                     //
//                  Global Public Licence (GPL)                        //
//                                                                     //
//                                                                     //
// This license described in this file overrides all other licenses    //
// that might be specified in other files for this library.            //
//                                                                     //
// This library is free software; you can redistribute it and/or       //
// modify it under the terms of the GNU Lesser General Public          //
// License as published by the Free Software Foundation; either        //
// version 2.1 of the License, or (at your option) any later version.  //
//                                                                     //
// This library is distributed in the hope that it will be useful,     //
// but WITHOUT ANY WARRANTY; without even the implied warranty of      //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU   //
// Library General Public License for more details.                    //
//                                                                     //
// You should have received a copy of the GNU Lesser General Public    //
// License along with this library (see COPYING.LIB); if not, write to //
// the Free Software Foundation :                                      //
//                      Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

}
program FIT_SHIP_DATA; 

%include 'fit_lsq_env.pas' {, list_on};


type
  ptr = ^rec;                   { Define a data record pointer }
  
  rec = record                  { * Define a data record }
    next:                  ptr; { Pointer to the next data record }
    speed,                      { Ship speed }
    o_fuel,                     { Used ship fuel }
    c_fuel:            lsq_flt  { Computed ship fuel }
  end;

var
  pymin, pymax, pyav3,          { Pointers to Minimum, Maximum and at middle range (in speed^3) fuel use record }
  frec, lrec:       ptr := nil; { Data record queue header pointer }

  sigmaf,                       { Standard error on the observed fuel use } 
  p0,  p1,  p2,                 { Parameters to fit and ... }
  sp0, sp1, sp2: lsq_flt :=0.0; { ... there related standard errors to fit }
  
  minx, maxx, miny, maxy,       { Mini-Maxi of speed and fuel use }
  mi0, mi1, mi2,
  ma0, ma1, ma2:       lsq_flt; { Mini-maxi of p0, p1, p2 }
  
  ndata,                        { Number of data }
  ncycle,                       { Current cycle number }
  nbcycle,                      { Required number of cycle }
  id0, id1, id2, ibl:  integer; { Index of each parameters }
  
  fit_name,                     { Name for this fit }
  inpfl, crvfl:         string; { Input, output and curve data file specification }

  inp, crv:               text; { Input and Output files }
  
  outlvl:              integer; { Output result level }
  bndfl, bmeafl:       boolean; { Flags }

{}


{  Define the access to the stadard CPASCAL run time messages (CPASCAL RTL routine) }
procedure PAS__GET_RTLMSG( n: integer; var msg: string ); external 'PAS__GET_RTLMSG';


procedure ERROR_AND_MSG( in_var msg: string; ierr: integer := -1 );
{ procedure to output any error message and stop if bstop is true }
var
  sysmsg: string;

begin
  WRITE( err, ' *** FitShip ' );
  if ierr <> 0 then WRITE( err, 'ERROR' )
               else WRITE( err, 'WARNING' );
  WRITELN( err, ' : ', msg );
  if ierr <> 0 then
  begin
    if ierr > 0 then
    begin
      PAS__GET_RTLMSG( ierr, sysmsg );
      WRITELN( err, ' RTL code ', iostatus:4, ' - msg : "', sysmsg, '"' )
    end;
    WRITELN( err, ' *** STOP FITSHIP ***' );
    PASCAL_EXIT( 2 )
  end
end ERROR_AND_MSG;



procedure GET_PARAM;
{ Use the command parameters as input and output file.
  Open its else Stop program on any error.
}
begin
  WRITELN;
  if argc < 2 then
  begin
    WRITE( ' FITSHIP Input file = ' ); READLN( inpfl )
  end
  else
  begin
    inpfl := argv[1]^;
    if (inpfl = '-h') or (inpfl = '-H') then
    begin
      WRITELN;
      WRITELN( ' Call with command :' );
      WRITELN( '    fitship <datafile>  <curve_file>' );
      WRITELN;
      WRITELN( ' to send the list and message output you must use the redirection as > output_listing' );
      WRITELN;
      WRITELN( ' Display this Help with the command fitship -h' );
      PASCAL_EXIT( 2 )
    end
  end;

  if argc < 3 then
  begin
    WRITE( ' FITSHIP curve output file = ' ); READLN( crvfl )
  end
  else crvfl := argv[2]^;
  
  OPEN( inp, inpfl, [read_file,error_file] );
  if iostatus <> 0 then ERROR_AND_MSG( 'Cannot open the input file', iostatus );

  OPEN( crv, crvfl, [write_file,error_file] );
  if iostatus <> 0 then ERROR_AND_MSG( 'Cannot open the output file', iostatus )
end GET_PARAM;



procedure READ_DATA;
{ Read the data from the input file }
var
  p, q:                    ptr;
  avx, dst, err:       lsq_flt;


  procedure RMEASURE;
  var
    sp, fu:    lsq_flt;

  begin
    if not EOF( inp) then READ( inp, sp );
    if not EOF( inp ) then
    begin
      READLN( inp, fu );
      ndata := ndata + 1;
      WRITELN( ' DATA line ', ndata:4, ' speed = ', sp:10:3, ', fuel = ', fu:10:3 )
    end;
    NEW( p );                           { Allocate a measure record and fill it }
    if pymin = nil then begin  pymin := p; pymax := p  end;
    with p^ do
    begin
      next   := nil;
      speed  :=  sp;
      o_fuel :=  fu;
      c_fuel := 0.0
    end;
    if frec = nil then
    begin
      minx := sp; maxx := sp;   { Initialize the Mini-Maxi }
      miny := fu; maxy := fu;
      frec := p                 { Append the measure record to the measure list }
    end
    else
    begin
      { Update the Mini-Maxi and related pointers as required }
      if minx > sp then minx := sp
                   else if maxx < sp then maxx := sp;
      if miny > fu then begin  miny := fu; pymin := p  end
                   else if maxy < fu then begin  maxy := fu; pymax := p  end;
      lrec^.next := p
    end;
    lrec := p
  end RMEASURE;


begin { READ_DATA }
  WRITELN( '  FITSHIP V1.0 of 12-OCT-2014' );
  WRITELN;
  READLN( inp, fit_name );              { Read the fit Name/Title/Reference }
  WRITELN( ' Title "', fit_name, '"' );
  READLN( inp, mi0, ma0 );              { Read the variation range of each parameters }
  WRITELN( ' P0 in [', mi0:10:3, ',', ma0:10:3, ']' );
  READLN( inp, mi1, ma1 );
  WRITELN( ' P1 in [', mi1:10:3, ',', ma1:10:3, ']' );
  READLN( inp, mi2, ma2 );
  WRITELN( ' P2 in [', mi2:10:3, ',', ma2:10:3, ']' );
  READLN( inp, sigmaf );                { Read the standard error on the Fuel use value }
  WRITELN( ' Common Std error (sigma) for all fuel use data = ', sigmaf:8:3 );
  READLN( inp, nbcycle );               { Read the required number of cycle(s) }
  WRITELN( ' Required cycle number = ', nbcycle:4 );
  ndata := 0;
  WRITELN;
  WRITELN( ' *** List of data ****' );
  WRITELN;
  RMEASURE;                             { Read the first measure }
  while not EOF( inp ) do
    RMEASURE;                           { Read the next measure or stop on EOF }
  
  WRITELN( ' <<< End Data >>>' );
  WRITELN;
  WRITELN;
  WRITELN;
  CLOSE( inp );                         { Close the data file }
  
  { Set the middle range measure record pointer }
  avx := (maxx*maxx*maxx + minx*minx*minx)*0.5;
  avx := EXP( LN( avx )/3.0 );          { Cubic root to compute avx when avx^3 = (maxx^3 + minx^3)/2 }
  err := (maxx - minx)*0.5;;
  p     := frec;
  pyav3 :=    p;
  while p <> nil do     { Loop to find the measure record with the nearest speed of avx }
  with p^ do
  begin
    dst := ABS( avx - speed );
    if dst < err then begin  pyav3 := p; err := dst end;
    p := next
  end
end READ_DATA;



procedure COMPUTE_FIRST_PARM_VALUES;
{ To set a first approximation of the parameters p0, P1 and P2 }
var
  x0, x1, x2, y0, y1, y2, ypa, ypb, f2, f0: lsq_flt;

begin
  p0 := miny;
  
  x0 := pymin^.speed; y0 := pymin^.o_fuel;
  x1 := pyav3^.speed; y1 := pyav3^.o_fuel;
  x2 := pymax^.speed; y2 := pymax^.o_fuel;
  x0 := x0*x0*x0*1.0e-6;
  x1 := x1*x1*x1*1.0e-6;
  x2 := x2*x2*x2*1.0e-6;
  (*
    WRITELN( ' * xy0 ', x0:8:1, y0:8:1, pymin^.speed:8:3 );
    WRITELN( ' * xy1 ', x1:8:1, y1:8:1, pyav3^.speed:8:3 );
    WRITELN( ' * xy2 ', x2:8:1, y2:8:1, pymax^.speed:8:3 );
  *)
  { Now the function is y = p0 + p1*exp( p2*x ) }
  ypa := (y1 - y0)/(x1 - x0);  { Get the two approximative primary derivates dy/dx }
  ypb := (y2 - y1)/(x2 - x1);
  (*
    WRITELN( ' * ypa = ', ypa:8:2, ', ypb = ', ypb:8:2 );
  *)
  { As dy/dx = p1*p2*exp( p2*x ) => ypb/ypa ~= exp( p2*xb - p2*xa )
    => p2 ~= LN( ypb/ypa )/(xb - xa)
    with xb - xa ~= (x2 - x0)/2 .
  }
  p2 := 2.0*LN( ypb/ypa )/(x2 - x0);
  (*
    WRITELN( ' * LN( ypb/ypa ) = ', LN( ypb/ypa ):8:4, ', p2 = ', p2:12:6 );
  *)
  { Now compute f0 = exp( p2*x0 ) and f2 = exp( p2*x2 ) because
    y2 - y0 = p1*(f2 - f0) => p1 = (y2 - y0)/(f2 - f0)
  }
  f0 := EXP( p2*x0 ); f2 := EXP( p2*x2 );
  p1 := (y2 - y0)/(f2 -f0);
  (*
    WRITELN( ' * f0, f2, p1 = ', f0:8:3, ' ', f2:8:3, ' ', p1:8:3 );
  *)
  p0 := ((y0 - p1*f0) + (y1 - p1*EXP( p2*x2 )) + (y2 - p1*f2))/3.0;
  (*
    WRITELN( ' * p0(0) = ', y0 - p1*f0:8:3, ', p0(2) = ', y2 - p1*f2:8:3, ', p0(1) = ', y1 - p1*EXP( p2*x2 ):8:3 );
  *)
  WRITELN;
  WRITELN( ' Initial set of deduced values for the three parameters :' );
  WRITELN( '    P0 = ', p0:12:5, ' P1 = ', p1:12:5, ',  P2 = ', p2:10:5 );
  WRITELN;
  WRITELN
end COMPUTE_FIRST_PARM_VALUES;



procedure SET_LEAST_SQUARES;
{ Initialize the Least-Squares Module and create the three parameters fit structures }
begin
  LSQ$INIT;                                     { Init the Least-Squares Module }
  id0 := LSQ$NEW_VPARM( p0, sp0, mi0, ma0 );    { Create the LSQ fitted variable p0 }
  id1 := LSQ$NEW_VPARM( p1, sp1, mi1, ma1 );    { Create the LSQ fitted variable p1 }
  id2 := LSQ$NEW_VPARM( p2, sp2, mi2, ma2 );    { Create the LSQ fitted variable p2 }
  ibl := LSQ$NEW_BLOCK( 1.0, 1.0 );             { Create our unique least-Square block }
  ncycle := 0
end SET_LEAST_SQUARES;



function  LSQ_CHANGE( i: integer; v_o, v_n, v_c, v_s, min, max: lsq_flt ): lsq_flt;
{ Procedure call by the Least-Squares Module (routine CYCLE_COMPLETE) to
  display the parameter change and apply some possible constraints
}
{ v_o is the old value of the current parameter,
  v_n is its new value,
  v_c is the related change,
  v_s is the related sigma (standard error),
  min and max are the allowed range.
}
var
  blm: boolean := true;
  
begin
  { Check for peak limit angles move inside the zone }

  if (v_n < min) then v_n := min
                 else if (v_n > max) then v_n := max
                                     else blm := false;

  { Output parameter Change when Required }
  WRITE( '    P', i-1:1, ' ':6, v_n:10:3, ' ', v_s:8:3, ' ', v_o:10:3, ' ', v_c:8:3 );
  if blm then WRITELN( ' Lm')
         else WRITELN;
  LSQ_CHANGE := v_n
end LSQ_CHANGE;



procedure FSHIP_CYCLE( bnd: boolean; var bchg: boolean );
{ Performs one cycle the Least-Squares (LSQ).

  bnd     is used to perform a final cycle without LSQ Variable change,
  bch     is used in output to signal an increasing of chi2,
  nobs will be the returned number of used observation point and bchg .
}
var
  p:                           ptr;
  x3, d1, d2, delta, desg: lsq_flt;
  nm:                      integer;

begin
  LSQ$INIT_CYCLE;                                       { Initialize the LSQ cycle with the Least-Squares Module }

  {             "_123/_123456.890__123456.890__123456.890___1234.678__1234.678 }
  WRITELN( '   #     Speed       Measure    Computed     delta   delta/sg' );
  WRITELN;

  nm := 0;
  p := frec;                                            { Init the loop on all measure record }
  while p <> nil do
  with p^ do
  begin
    LSQ$INIT_DERV;                                      { Init the table of partial derivates in the Least-Squares Module }
    { Now compute the Fuel use and its partial derivates }
    { Formula c_fuel = p0 + p1*EXP( p2*(speed/100)^3 )  =>  y = p0 + p1*EXP( p2*x^3 ) }
    x3 := speed*speed*speed*1.0e-6;
    d1 := EXP( p2*x3 );
    d2 := p1*x3*d1;
    c_fuel := p0 + p1*d1;
    LSQ$SET_DERV( id0, 1.0 );                           { Send the three partial derivates to the Least-Squares Module }
    LSQ$SET_DERV( id1, d1 );
    LSQ$SET_DERV( id2, d2 );
    delta := o_fuel - c_fuel;
    LSQ$SET_CONTR( delta, 1.0/sigmaf );                 { Send the difference measure-computed value to the Least-Squares Module }
                                                        { Note: the used weight (by the Least-Squares Module) will be (1/sigmaf)^2 }

    { We can output here the results for this measure }
    nm := nm + 1;
    WRITELN( ' ', nm:3, '/ ', speed:10:3, '  ', o_fuel:10:3, '  ', c_fuel:10:3, '  ', delta:8:3, ' ', delta/sigmaf:8:3 );
 (* output for debug :
    WRITELN( ' * V^3 = ', x3:8:1, ' EXP = ', d1:8:1, ', p1*EXP = ', d2:8:1 );
  *)
    p := next
  end;

  { Perform the Least-Squares matrix computing }
  
  if not bnd then
  begin
    WRITELN;
    WRITELN;
    WRITELN( ' Cycle # ', ncycle:3, ' Results' );
    WRITELN;
    {         ____P1______123456.890__1234.890__123456.678__1234.678 }
    WRITELN( ' parameter   New_Value    Sigma    Old_Valuet   Change'  );
    WRITELN
  end;
 
  bchg := LSQ$PROCESS( bnd, not bnd, true, LSQ_CHANGE );
  
  WRITELN;
  WRITELN( ' Least_Squares Goodness of fit (chi2) = ', chi2:8:2 );
  WRITELN;
  WRITELN;
  WRITELN
end FSHIP_CYCLE;



procedure FSHIP_PROCESS;
{ Perform a complete Least-Squares fit.
}
const
  eps = 1.0e-4;                         { Epsilon for convergence test }
  mfc =   10.0;                         { Levenberg-Marquardt change factor }

var
  ik0, ik1, ndata:     integer;
  pmrq, lmrq, mcoef:   lsq_flt;
  bch, bok:            boolean;

begin
  bok   :=    true;
  mcoef :=     1.0;                     { Set the start Levenberg-Marquardt factor change }
  lmrq  :=     1.0;                     { Initialize the Levenberg-Marquardt factor }
  ncycle := 0; ik0 := 0; ik1 := 0;
  repeat
    ncycle   := ncycle + 1;
    pmrq := lmrq;
    lmrq := 1.0/(1.0 + mcoef);
    LSQ$SET_DMP_MRQ( 1, {dmp_r} lmrq, lmrq );

    { Output of cycle and computed measures titles lines }
    WRITELN;
    WRITELN;
    WRITELN( ' *** Cycle # ', ncycle:3, ' with mrq = ', lmrq:8:4, ' ***' );
    {               *** Cycle # 123 with mrq = 123.5678 *** }
    WRITELN( ' ---------------------------------------' );
    WRITELN;

    FSHIP_CYCLE( false, bch );          { Perform the cycle computing }
    if not bch then
    begin
      ik1 := ik1 + 1;
      mcoef := mcoef*mfc;
      if ncycle > 1 then
      begin
        WRITELN( ' !! Cycle # ', ncycle:3, ' Canceled (Too large resulting chi2 : ', chi2:10:2, ' )' );
        WRITELN( ' ==> Retry with a decreasing Levenberg-Marquardt factor.' );
        WRITELN
      end;
      LSQ$RESTORE_PARM;                 { Restore the previuos value of parameters to fit }
      bch := true;                      { New activation of parameter change }
      FSHIP_CYCLE( false, bch );        { Repeat to Perform the cycle computing }
      ik0 := 0; bok := false
    end
    else
    begin
      if ABS( lchi2 - chi2 ) <= chi2eps then ik0 := ik0 + 1
                                        else ik0 := 0;
      if bok then mcoef := mcoef/mfc;
      ik1 := 0;
      (*
        if ncycle > 1 then WRITELN( ' Cycle # ', ncycle:3, ' : Goodness of fit (Chi2) = ', chi2:10:2 );
      *)
      bok := true
    end
  until (ncycle >= nbcycle) or (ik0 > 1);
  LSQ$SET_DMP_MRQ( 1, 1.0, 1.0 );
  WRITELN;
  if ik0 <= 1 then
  begin
    WRITELN;
    WRITE( '  *** Stop the fit after ', nbcycle:0, ' Least-Squares cycle' );
    if nbcycle >1 then WRITE( 's' );
    WRITELN( ' without complete convergence ***' );
    WRITELN
  end;
  WRITELN;
  WRITELN( '     *** ***   Final  Cycle   *** ***' );
  WRITELN( '     --------------------------------' );
  WRITELN;
  FSHIP_CYCLE( true, bch )              { Perform the final cycle with computing of correlation matrix }
end FSHIP_PROCESS;


procedure WRITE_RESULT;
var
  rvl:           lsq_flt;
  title:          string;

begin
  WRITELN;
  WRITELN;
  WRITEV( title,'*** Ship Data references "', fit_name, '" ***' );
  WRITELN( ' ', title );
  for ij := 1 to title.length do  title[ij] := '-';
  WRITELN( ' ', title );
  WRITELN;
  WRITELN;
  WRITELN( ' *** Final Fit Result ***' );
  WRITELN( ' ------------------------' );
  WRITELN;
  WRITELN( '  The used formula is : Fu = P0 + P1*EXP( P2*(speed/100)^3 )' );
  WRITELN;
  WRITELN;
  WRITELN( '   Final goodness of fit (chi2) = ', chi2:10:3 );
  WRITELN;
  {              ___________123456.890__123456.890 }
  WRITELN( ' Parameter     value       sigma ' );
  WRITELN( ' --------------------------------' );
  WRITELN;
  WRITELN( '    P0  ', p0:10:3, '  ', sp0:10:3 );
  WRITELN( '    P1  ', p1:10:3, '  ', sp1:10:3 );
  WRITELN( '    P2  ', p2:10:3, '  ', sp2:10:3 );
  WRITELN;
  WRITELN;
  WRITELN( '    The Correlation matrix is :' );
  WRITELN;
  {         ss______12.3456__12.3456__12.3456 }
  WRITELN( '          P0       P1       P2' );
  WRITELN;
  for i := 1 to 3 do
  begin
    WRITE( '  P', i-1:1, '  ' );
    for j := 1 to 3 do
      if j < i then WRITE( ' ':8 )
               else
                 if j = i then WRITE( '  1.0000' )
                          else
                          begin
                            rvl := LSQ$CORREL( i, j );
                            WRITE( '  ', rvl:6:4 )
                          end;

    WRITELN
  end;
  WRITELN;
  WRITELN
end WRITE_RESULT;



procedure WRITE_CURVE_TABLE;
const
  ncpt = 200; { Number of points for the computed curve }

var
  p:                               ptr;
  rng, stp, xx, yy:            lsq_flt;
  xmi, xma, ymi, yma:          lsq_flt;
  xytab:array[1..ncpt,1..2] of lsq_flt;

begin
  WRITELN( crv, ' FITSHIP  CURVE' );
  WRITELN( crv, ' ', fit_name );
  WRITELN( crv, ' ', p0:10:3, ' ', p1:10:3, ' ', p2:10:3 );
  WRITELN( crv, ' ', sp0:10:3, ' ', sp1:10:3, ' ', sp2: 10:3 );

  { Size the graphic space required for observed points and computed curve }
  rng := maxx - minx;
  xmi := minx -0.1*rng; xma := maxx + 0.1*rng;
  rng := xma - xmi; stp := rng/(ncpt - 1);
  ymi := frec^.o_fuel; yma := ymi;

  { Create the computed curve points list }
  xx := xmi;
  for ii := 0 to ncpt-1 do
  begin
    xx := (xmi + ii*stp);
    yy := p0 + p1*EXP( p2*xx*xx*xx*1.0e-6 );
    if yy < ymi then ymi := yy
                else if yy > yma then yma := yy;
    xytab[ii+1,1] := xx; xytab[ii+1,2] := yy
  end;

  { Output the required range for x and y plot axis }
  WRITELN( crv, ' ', xmi:6:1, ' ', xma:6:1, ' ', ymi:6:0, ' ', yma:6:0, '             # Unit range for each plot axis.' );
  
  { Output the observed points list }
  WRITELN( crv, ndata: 5 );
  p := frec;
  while p <> nil do
  with p^ do
  begin
    WRITELN( crv, ' ', speed:10:3, ' ', o_fuel:10:3 );
    p := next
  end;

  { Output the computed curve points list }
  WRITELN( crv, ncpt:5 );
  for ii := 1 to ncpt do
    WRITELN( crv, ' ', xytab[ii,1]:10:3, ' ', xytab[ii,2]:10:3 );

  CLOSE( crv )
end WRITE_CURVE_TABLE;



begin { Main Program }
  GET_PARAM;
  READ_DATA;
  COMPUTE_FIRST_PARM_VALUES;
  SET_LEAST_SQUARES;
  FSHIP_PROCESS;
  WRITE_RESULT;
  WRITE_CURVE_TABLE;
  WRITELN( ' Normal End.' )
end FIT_SHIP_DATA.

