{
*************************************************************************
*                                                                       *
*                                                                       *
*                    F I T S H I P -  D R A W                           *
*                                                                       *
*                                                                       *
*                Version  1.0-A  - - 16-Oct-2014                        *
*                                                                       *
*                                by                                     *
*                                                                       *
*                  Pierre Wolfers, Institut Neel                        *
*                                                                       *
*          CNRS GRENOBLE,  25 Avenue des Martyrs, B.P. 166              *
*                                                                       *
*                     F 38042 GRENOBLE CEDEX 9                          *
*                                                                       *
*                           F R A N C E                                 *
*                                                                       *
*                                                                       *
*                                                                       *
*************************************************************************

/////////////////////////////////////////////////////////////////////////
//                                                                     //
//                                                                     //
//                  Global Public Licence (GPL)                        //
//                                                                     //
//                                                                     //
// This license described in this file overrides all other licenses    //
// that might be specified in other files for this library.            //
//                                                                     //
// This library is free software; you can redistribute it and/or       //
// modify it under the terms of the GNU Lesser General Public          //
// License as published by the Free Software Foundation; either        //
// version 2.1 of the License, or (at your option) any later version.  //
//                                                                     //
// This library is distributed in the hope that it will be useful,     //
// but WITHOUT ANY WARRANTY; without even the implied warranty of      //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU   //
// Library General Public License for more details.                    //
//                                                                     //
// You should have received a copy of the GNU Lesser General Public    //
// License along with this library (see COPYING.LIB); if not, write to //
// the Free Software Foundation :                                      //
//                      Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  //
//                                                                     //
/////////////////////////////////////////////////////////////////////////

}
program FITSHIP_DRAW;


%include 'PASENV:draw_defs.pas';        { Get the DRAW environment }

const
  easybox = true;                       { Select the type ofaxis box }

type
  intarr2 = array[1..2] of integer;
  intarr4 = array[1..4] of integer;
  fltarr2 = array[1..2] of  Dfloat;
  fltarr4 = array[1..4] of  Dfloat;

const
  max_obs =  256;                       { Maximum number of point for the experimental curve }
  max_cpt = 1024;                       { Maximum number of point for the computed curve }

  untx =  'Knots';                      { Unit string for X and Y axis }
  unty = 'KGal/h';

  xa_dir = Dpoint3[  1.0,  0.0,  0.0 ]; { Directions vectors for the two axis and there tics }
  xt_dir = Dpoint3[  0.0,  1.0,  0.0 ];
  ya_dir = Dpoint3[  0.0,  1.0,  0.0 ];
  yt_dir = Dpoint3[ -1.0,  0.0,  0.0 ];

  ax_tfre = intarr4[ 18,  9,  1,  0 ];
  ay_tfre = intarr4[  5,  1,  0,  0 ];
  ax_tknd = intarr4[  1,  1,  1,  0 ];
  ay_tknd = intarr4[  1,  1,  0,  0 ];
  ax_tsiz = fltarr4[ 0.6, 0.4, 0.2, 0.0 ];
  ay_tsiz = fltarr4[ 0.6, 0.3, 0.0, 0.0 ];

  bx_dir = intarr2[ Axis_PFlg_YPos + Axis_PFlg_TSym +   { Axis with Y shift specified at y axis value }
                    Axis_PFlg_TVl  + Axis_Plot_X,       { symmetrized ticks, Tick Values, X axis at y = 0 }
                    Axis_PFlg_TVl  + Axis_Plot_Y ];     { Tick Values, L./R. Permuttation }
  bx_shift =  fltarr2[0.0, 0.0];        {  }

 
type
  cvxy = record                         { Define the curve point }
    xx, yy: Dfloat
  end;

var
  mline,                                { String to edit any message }
  inpfl:                        string; { Input File specification }

  fitname:                      string; { Reference of the fit (to use as plot title) }

  p0, sp0, p1, sp1, p2, sp2,            { Fitted Parameters and related std. error (sigmas) }
  mix, max, miy, may:           Dfloat; { Mini-Maxi of value for each plot axis }

  n_obs, n_cpt:                integer; { Number of point for eavh curve }


  obs_tab: array[1..max_obs] of cvxy;   { Tables for experimental and computed curve points }
  cpt_tab: array[1..max_cpt] of cvxy;

  inp:                            text; { Input file as created by FITSHIP }


  unt:                       unit_type; { Unity Type (raster/cm for 0/1) }
  bstatus:                     boolean; { Status word for DRAW procedure return }
  box,                                  { Axis box identifier }
  x_axis, y_axis,                       { AXis identifier }
  ireq:                        integer; { Request number as returned by DRAW$DIALOG on user action }
  xs, ys,                               { x,y Sizes of the canvas to plot in the unt units }
  r:                            Dfloat;


{}

  
  status, segb1, iseg1, iseg2, iseg3, iseg4: Dint;







{  Define the access to the stadard CPASCAL run time messages (CPASCAL RTL routine) }
procedure PAS__GET_RTLMSG( n: integer; var msg: string ); external 'PAS__GET_RTLMSG';


procedure ERROR_AND_MSG( in_var msg: string; ierr: integer := -1 );
{ Procedure to output any error message and stop if bstop is true }
var
  sysmsg: string;

begin
  WRITE( err, ' *** FitShip ' );
  if ierr <> 0 then WRITE( err, 'ERROR' )
               else WRITE( err, 'WARNING' );
  WRITELN( err, ' : ', msg );
  if ierr <> 0 then
  begin
    if ierr > 0 then
    begin
      PAS__GET_RTLMSG( ierr, sysmsg );
      WRITELN( err, ' RTL code ', iostatus:4, ' - msg : "', sysmsg, '"' )
    end;
    WRITELN( err, ' *** STOP FITSHIP ***' );

    PASCAL_EXIT( 2 )
  end
end ERROR_AND_MSG;



procedure GET_PARAM;
{ Procedure to get the program argument and open the input file }
begin
  WRITELN;
  if argc < 2 then
  repeat
    WRITE( ' Curve file as created by FitShip program = ' );
    READLN( inpfl )
  until inpfl.length > 0
  else inpfl := argv[1]^;
  OPEN( inp, inpfl, [read_file,error_file] );
  if iostatus <> 0 then ERROR_AND_MSG( 'Cannot open the input file', iostatus )
end GET_PARAM;



procedure STR_CLEAN( var str: string );
{ Procedure to suppress first and last realining space from a string }
var
  i, j, l: integer;
  ch: char;
  bs: boolean;
  
begin
  if str.length > 0 then
  begin
    i := 1; bs := true;
    while bs and (i <= str.length) do
      if str[i] <= ' ' then i := i + 1
                       else bs := false;
    j := str.length; bs := true;
    while bs and (j > 0) do
      if str[j] <= ' ' then j := j - 1
                       else bs := false;
    str := SUBSTR( str, i, j-I+1 )
  end
end STR_CLEAN;



procedure GET_DATA;
var
  stline:               string;
  ch:                     char;
  bok:                 boolean;
  i,j:                 integer;

begin
  READLN( inp, stline );
  if stline <> ' FITSHIP  CURVE' then ERROR_AND_MSG( 'Bad input file (Incorrect header identifier)', -1 );
  READLN( inp, fitname );               { Read the Fit references ... }
  STR_CLEAN( fitname );                 { ... and clean it to can use as plot title }
  READLN( inp, p0, p1, p2 );            { Read the Fitted parameter values and related sigma }
  READLN( inp,  sp0, sp1, sp2 );
  READLN( inp, mix, max, miy, may );    { Readconst
  easybox = false;

 the minimum and maximum values used for each plot axis }

  { Now we read the experimental Curve data }
  READLN( inp, n_obs );                 { Read the number of point to read }
  if n_obs > max_obs then
  begin
    WRITEV( stline, 'Too many point in the experimental curve(maximum is ', max_obs:0, ')' );
    ERROR_AND_MSG( stline, -2 )
  end;
  for i := 1 to n_obs do
    with obs_tab[i] do
      READLN( inp, xx, yy );

  { Now we read the computed Curve data }
  READLN( inp, n_cpt );                 { Read the number of point to read }
  if n_obs > max_cpt then
  begin
    WRITEV( stline, 'Too many point in the experimental curve(maximum is ', max_obs:0, ')' );
    ERROR_AND_MSG( stline, -2 )
  end;
  for i := 1 to n_cpt do
    with cpt_tab[i] do
      READLN( inp, xx, yy );
 
  CLOSE( inp )
end GET_DATA;



procedure INIT_PLOT;
begin
  Draw_Server_Mode := DrwSrv_Disable_IniErr;    { Set the Init Error User management DRAW Mode }
  DRAW$INIT( xs, ys, unt, 'FitShip PLOT' );     { Build the Plot Window and get the canvas sizes }
  { Try to init the DRAW system. When the Draw Server flags word is 0 => The init fails }
  if Draw_Server_Flags = 0 then ERROR_AND_MSG( 'Cannot Start the DRAW_SERVER (DRAW not installed ?)', -99 );

  { DRAW$INIT OK: xs,ys are the sizes of drawable window }

  DRAW$PIC_VIEW( 1 );                           { Set to graphic update as soon as mode }
  bstatus := DRAW$PICTURE( fitname, 26.0, 20.0, true, true );   { Build the PLOT SPACE with a space of 25 cm * 20 cm }

end INIT_PLOT;


procedure PLOT_CURVES;
var
  bpen: boolean;

begin
  { Create the axis box }
  DRAW$NEW_SEG( 1 );                    { Create a specific segment for the axis bok }

  if easybox then
    box := DRAW$EASY_BOX_2D(  2.0,  2.0,       { Coordinates of the box origine (low,left origine) }     
                             22.0, 17.0,       { Size of the X and Y axis }
                              0.0,  max,       { Minimum and maximum of speed values along X axis }
                              0.0,  may,       { Minimum and maximum of fuel consumption values along Y axis }
                              untx,            { Unit name for ship speed - X axis }
                              unty,            { Unit name for fuel consumption - Y axis }
                              0                { Coordinates schema number }
                         )
  else
  begin
    x_axis := DRAW$NEW_AXIS( xa_dir, xt_dir, 22.0, 0.0, max, 0, Axis_Arrow+Axis_ArroF );
    y_axis := DRAW$NEW_AXIS( ya_dir, yt_dir, 17.0, 0.0, may, 0, Axis_Arrow+Axis_ArroF+Axis_VOther );

    Draw$Set_Axis_Unit( x_axis, untx );   { Set the unit for each axis }
    Draw$Set_Axis_Unit( y_axis, unty );

    Draw$Set_Axis_Ticks( x_axis,  0,  3, ax_tfre, ax_tknd, ax_tsiz );
    Draw$Set_Axis_Ticks( y_axis,  0,  2, ay_tfre, ay_tknd, ay_tsiz );

    Draw$Set_Axis_Value( x_axis, 2, 112,,,,, 4,  4, 0 ); { Set Value Font, Field and decimal }
    Draw$Set_Axis_Value( y_axis, 2, 112,,,,, 4,  6, 0 ); { Set Value Font, Field and decimal }

    box := DRAW$NEW_BOX( Dpoint3[2.0, 2.0, 0.0], x_axis, y_axis, 0, bx_dir, bx_shift );

  end;
  
  DRAW$PLOT_BOX( box );                 { Display the axis box }
  DRAW$SEG_END;                         { Close the axis box }
  
  DRAW$OPEN_BOX( box );                 { Open the box to plot inside }

  DRAW$NEW_SEG( 2 );                    { Open the observed curve segment }
  DRAW$SEG_ATTR( 2, 1, 0, 1, 0.0 );     { Set it as detectable, and visible }
  DRAW$OUT_MODE( PLOT_MARKER );         { select the Marker mode }
  DRAW$MARKER_ATTR( 2, 1.5 );           { Select the single cross of the size 1.5 }
  with obs_tab[1] do
    DRAW$PLOT( xx, yy, false );         { Goto the first point pen up }
  for ii := 2 to n_obs do               { Loop to plot all other points of the curve }
    with obs_tab[ii] do
      DRAW$PLOT( xx, yy, true );
  DRAW$RPLOT( 0.0, 0.0, false );        { Set pen up }
  DRAW$SEG_END;

  WRITELN( ' *** cpt = ', n_cpt );
  DRAW$NEW_SEG( 3 );                    { Open the computed curve segment }
  DRAW$SEG_ATTR( 3, 1, 0, 1, 0.0 );     { Set it as detectable, and visible }
  DRAW$OUT_MODE( PLOT_LINE );           { Select the Line mode }
  DRAW$LINE_ATTR( 1, 1.5 );             { Select a continue line of the thickness of 1.5 }
  with cpt_tab[1] do
    DRAW$PLOT( xx, yy, false );         { Goto the first point pen up }
  for ii := 2 to n_cpt do               { Loop to plot all other points of the curve }
    with cpt_tab[ii] do
      DRAW$PLOT( xx, yy, true );
  DRAW$RPLOT( 0.0, 0.0, false );        { Set pen up } 
  DRAW$SEG_END;
  
end PLOT_CURVES;



begin { Main }
  GET_PARAM;
  GET_DATA;

  INIT_PLOT;

  PLOT_CURVES;

  repeat
    ireq := DRAW$DIALOG( 1 );                   {  }
  
  until ireq < 0;
  DRAW$END
end FITSHIP_DRAW.

